from .helpers.confchecks import (
    check_function, CheckResult, action_function, ActionResult
)
import json
from datetime import datetime, timedelta
from dcicutils import s3_utils
from dcicutils.ff_utils import get_metadata, search_metadata, patch_metadata, post_metadata


@check_function(action="refresh_access_keys")
def access_key_status(connection, **kwargs):
    """ Checks the creation date of the admin access keys and if expiration is soon
        emit warning and error eventually as it gets closer. Once these keys expire
        (every 90 days), foursight and tibanna will no longer function.
    """
    check = CheckResult(connection, 'access_key_status')
    check.action = 'refresh_access_keys'
    # TOOD: Figure this out ... Seems like, both from this code and what we are seeing in the actual action history,
    # that the refresh action is running everday; we only want to run a refresh if the access is expiring very soon.
    # Always allow refresh
    # TODO/INPROGRESS/2023-06-16/dmichaels:
    # But I think if we do not want the action to run automatically then set check.action to "" (cannot be None). 
    # No nevermind that does not work because then no action is shown at all.
    # Try setting allow_action to False ...
    # Works but then not always enabled ... need another bit ...
    check.allow_action = True
    fs_user_email, fs_user_kp = 'foursight.app@gmail.com', 'access_key_foursight'
    user_props = get_metadata(f'/users/{fs_user_email}?datastore=database', key=connection.ff_keys)
    user_uuid = user_props['uuid']
    access_keys = search_metadata(f'/search/?type=AccessKey&description={fs_user_kp}&user.uuid={user_uuid}'
                                  f'&sort=-date_created', key=connection.ff_keys)
    most_recent_key = access_keys[0]  # should always be present if deploy has run
    # date format: 2022-07-05T01:01:43.498347+00:00 (isoformat)
    most_recent_key_creation_date = datetime.fromisoformat(most_recent_key['date_created'])
    expiration_date = most_recent_key_creation_date + timedelta(days=90)
    one_week_to_expiration = expiration_date - timedelta(days=7)
    three_weeks_to_expiration = expiration_date - timedelta(days=21)
    now = datetime.now(most_recent_key_creation_date.tzinfo)
    if now > one_week_to_expiration:
        check.status = 'FAIL'
        check.summary = (f'Application access keys will expire in less than 7 days! Please run'
                         f' the deployment action ASAP')
        check.brief_output = check.full_output = check.summary
        return check
    elif now > three_weeks_to_expiration:
        check.status = 'WARN'
        check.summary = (f'Application access keys will expire in less than 21 days! Please run'
                         f' the deployment action soon')
        check.brief_output = check.full_output = check.summary
        check.prevent_action = True
        return check
    else:
        check.status = 'PASS'
        check.summary = (f'Application access keys expiration is more than 3 weeks away. All good.'
                         f' Expiration date: {expiration_date}')
        check.prevent_action = True
        return check


@action_function()
def refresh_access_keys(connection, **kwargs):
    """ Triggers a refresh of the 3 admin keys, previously run through the portal """
    action = ActionResult(connection, 'refresh_access_keys')
    admin_keys = [('4dndcic@gmail.com', 'access_key_admin'),  # fourfront admin
                  ('cgap.platform@gmail.com', 'access_key_admin'),  # cgap admin
                  ('tibanna.app@gmail.com', 'access_key_tibanna'),
                  ('foursight.app@gmail.com', 'access_key_foursight')]
    s3 = s3_utils.s3Utils(env=connection.ff_env)
    full_output = {
        'successfully_generated': []
    }
    for email, kp_name in admin_keys:
        try:
            user = get_metadata(f'/users/{email}?datastore=database', key=connection.ff_keys)
        except Exception:
            continue  # user not found
        user_uuid = user['uuid']
        access_keys = search_metadata(f'/search/?type=AccessKey&description={kp_name}&user.uuid={user_uuid}'
                                      f'&sort=-date_created', key=connection.ff_keys)
        # generate new key
        access_key_req = {'user': user_uuid, 'description': kp_name}
        # 2020-06-13/dmichaels: The actual result returned by the portal for this POST is not what
        # seems to be expected; the access_key_id and secret_access_key are not within the @graph
        # array; but handle both cases just in case; maybe that as an older (or newer) API.
        # access_key_res = post_metadata(access_key_req, 'access-keys', key=connection.ff_keys)['@graph'][0]
        access_key_res = post_metadata(access_key_req, 'access-keys', key=connection.ff_keys)
        access_key_id = access_key_res.get('access_key_id')
        secret_access_key = access_key_res.get('secret_access_key')
        if not access_key_id or not secret_access_key:
            # We will say these must occur in pairs; both at the top level or both within the @graph array.
            graph_item = access_key_res.get('@graph', [{}])[0]
            access_key_id = graph_item.get('access_key_id')
            secret_access_key = graph_item.get('secret_access_key')
        s3_obj = {'secret': secret_access_key, 'key': access_key_id, 'server': s3.url}
        s3.s3_put_secret(json.dumps(s3_obj), kp_name)
        full_output['successfully_generated'].append(email)
        # clear out old keys after generating new one
        for access_key in access_keys:  # note this search result was computed before the new key was added
            if access_key['status'] != 'deleted':
                try:
                    patch_metadata(patch_item={'status': 'deleted'}, obj_id=access_key['uuid'], key=connection.ff_keys)
                except Exception as e:
                    print(f"Exception while trying to delete old access key ({access_key['uuid']})")
                    print(e)

    action.full_output = full_output
    action.status = 'DONE'
    return action



import logging
import os

from dcicutils.cloudformation_utils import AbstractOrchestrationManager


logging.basicConfig()
logger = logging.getLogger(__name__)


class Stage(object):

    prod_stage_name = 'prod'

    def __init__(self, foursight_prefix):
        self.prefix = foursight_prefix

    @classmethod
    def get_stage_from_env_variable(cls):
        # set environmental variables in .chalice/config.json
        return os.environ.get('chalice_stage', 'dev')  # default to dev

    @classmethod
    def get_stage(cls):
        stage = cls.get_stage_from_env_variable()
        if stage == 'test':
            stage = 'dev'
        return stage

    def get_queue_name(self):
        return '-'.join([self.prefix, self.get_stage_from_env_variable(), 'check_queue'])

    ENCACHE_RUNNER_NAME = True

    def get_runner_name(self, encache=ENCACHE_RUNNER_NAME):
        """
        Gets the name of the Lambda function to use as a check runner.
        """
        stage = self.get_stage()
        check_runner = os.environ.get('CHECK_RUNNER', None)

        if not check_runner:
            check_runner = AbstractOrchestrationManager.discover_foursight_check_runner_name(stage=stage,
                                                                                             encache=encache)

        if not check_runner:
            # TODO: Is there EVER a situation where this works any more? Can we ditch this heuristic? -kmp 23-Jun-2022
            check_runner = '-'.join([self.prefix, stage, 'check_runner'])

        return check_runner

    @classmethod
    def is_stage_prod(cls):
        if cls.get_stage() == cls.prod_stage_name:
            return True
        else:
            return False



import boto3
from .datetime_utils import convert_utc_datetime_to_useastern_datetime_string
from .yaml_utils import load_yaml
from collections import OrderedDict
from dcicutils.function_cache_decorator import function_cache
from dcicutils.obfuscation_utils import obfuscate_dict
from typing import Optional, Union


_STACK_NAME_PREFIX = "c4-"


@function_cache
def aws_get_stacks() -> list:
    """
    Returns the list of all known AWS CloudFormation stacks with various metadata.
    """
    def stack_id(element):
        return element.name
    stacks_info = []
    cf = boto3.resource("cloudformation")
    for stack in sorted(cf.stacks.all(), key=stack_id):
        if stack.name.startswith(_STACK_NAME_PREFIX):
            stacks_info.append(_create_aws_stack_info(stack))
    return stacks_info


def _create_aws_stack_info(stack: object):
    """
    Returns the given boto3 format AWS stack info in our canonical form.
    """
    return {
        "name": stack.name,
        "id": stack.stack_id,
        "description": stack.description,
        "role_arn": stack.role_arn,
        "status": stack.stack_status,
        "updated": convert_utc_datetime_to_useastern_datetime_string(stack.last_updated_time),
        "created": convert_utc_datetime_to_useastern_datetime_string(stack.creation_time)
    }


@function_cache
def aws_get_stack(stack_name_or_object: Union[str, object]) -> dict:
    """
    Returns all detailed info for the given AWS CloudFormation stack name,
    including various metadata, outputs, parameters, and resources.
    Output and parameter values are obfuscated if their names represents a sensitive value.
    """
    result = {}
    stack = _get_aws_stack_object(stack_name_or_object)
    if stack:
        result = _create_aws_stack_info(stack)
        outputs = aws_get_stack_outputs(stack)
        parameters = aws_get_stack_parameters(stack)
        resources = aws_get_stack_resources(stack)
        result["outputs"] = outputs
        result["parameters"] = parameters
        result["resources"] = resources
    return result


@function_cache
def aws_get_stack_outputs(stack_name_or_object: Union[str, object]) -> dict:
    """
    Returns the name/value outputs for the given AWS CloudFormation stack name.
    Output values are obfuscated if the output name represents a sensitive value.
    """
    def output_id(element):
        return element["OutputKey"]
    result = {}
    stack = _get_aws_stack_object(stack_name_or_object)
    if stack and stack.outputs:
        for stack_output in sorted(stack.outputs, key=output_id):
            result[stack_output.get("OutputKey")] = stack_output.get("OutputValue")
    return _obfuscate(result)


@function_cache
def aws_get_stack_parameters(stack_name_or_object: Union[str, object]) -> dict:
    """
    Returns a name/value dictionary of the parameters for the given AWS CloudFormation stack name.
    Parameter values are obfuscated if the parameter name represents a sensitive value.
    """
    def parameter_id(element):
        return element["ParameterKey"]
    result = {}
    stack = _get_aws_stack_object(stack_name_or_object)
    if stack and stack.parameters:
        for stack_parameter in sorted(stack.parameters, key=parameter_id):
            result[stack_parameter.get("ParameterKey")] = stack_parameter.get("ParameterValue")
    return _obfuscate(result)


@function_cache
def aws_get_stack_resources(stack_name_or_object: Union[str, object]) -> dict:
    """
    Returns a name/value dictionary of the resources for the given AWS CloudFormation stack name.
    """
    def resource_id(element):
        return element.logical_resource_id
    result = {}
    stack = _get_aws_stack_object(stack_name_or_object)
    if stack:
        for stack_resource in sorted(list(stack.resource_summaries.all()), key=resource_id):
            result[stack_resource.logical_resource_id] = stack_resource.resource_type
    return _obfuscate(result)


def _get_aws_stack_object(stack_name_or_object: Union[str, object]) -> Optional[object]:
    """
    Returns the boto3 Cloudformation stack object for the given stack name, if the given argument
    is a string, or if the given argument is already a boto3 stack object, then return that value.
    :param stack_name_or_object: String representing Cloudformation stack name or boto3 stack object.
    :returns: The boto3 Cloudformation stack object or None if not found.
    """
    if type(stack_name_or_object).__name__ == "cloudformation.Stack":
        return stack_name_or_object
    elif not isinstance(stack_name_or_object, str):
        return None
    cf = boto3.resource("cloudformation")
    stack = list(cf.stacks.filter(StackName=stack_name_or_object))
    return get_single_item_list_value(stack)
    # return stack[0] if len(stack) == 1 else None


def get_single_item_list_value(any_list: list) -> object:
    if len(any_list) != 1:
        raise Exception(f"Single item list expected but contains {len(any_list)}")
    return any_list[0]


@function_cache
def aws_get_stack_template(stack_name: str) -> dict:
    """
    Returns the AWS Cloudformation template as a dictionary for the given stack name.
    The entire template will be obfuscated according to the obfuscate_dict function.
    """
    cf = boto3.client("cloudformation")
    stack_template = cf.get_template(StackName=stack_name)
    stack_template_body = stack_template["TemplateBody"]
    if isinstance(stack_template_body, dict) or isinstance(stack_template_body, OrderedDict):
        #
        # For some reason for our AWS stack c4-foursight-cgap-supertest-stack
        # in particular, we get back an OrderedDict rather than the usual string.
        #
        stack_template_dict = stack_template_body
    else:
        #
        # For other AWS stacks like c4-datastore-cgap-supertest-stack,
        # we get a simple string containing the YAML for the template.
        #
        stack_template_dict = load_yaml(stack_template_body)
    return _obfuscate(stack_template_dict)


def _obfuscate(value: dict) -> dict:
    return obfuscate_dict(value, obfuscated="********")



class BadCheckOrAction(Exception):
    """
    Generic exception for a badly written check or library.
    __init__ takes some string error message
    """
    def __init__(self, message=None):
        # default error message if none provided
        if message is None:
            message = "Check or action function seems to be malformed."
        super().__init__(message)


class BadCheckSetup(Exception):
    """
    Generic exception for an issue with a check setup.
    __init__ takes some string error message
    """
    def __init__(self, message=None):
        # default error message if none provided
        if message is None:
            message = "Malformed check setup found."
        super().__init__(message)


class MissingFoursightPrefixException(Exception):
    """
    Generic exception for an issue with foursight prefix
    not defined or initialized before using a method that
    requires it.
    __init__ takes some string error message
    """
    def __init__(self, message=None):
        # default error message if none provided
        if message is None:
            message = "Foursight prefix is missing."
        super().__init__(message)



import datetime
import time
import logging
from dateutil import tz
from abc import abstractmethod
import json
# from foursight_core.s3_connection import S3Connection
from foursight_core.exceptions import (
  BadCheckOrAction,
  MissingFoursightPrefixException
)


logging.basicConfig()
logger = logging.getLogger(__name__)


class RunResult(object):
    """
    Generic class for CheckResult and ActionResult. Contains methods common
    to both.
    """
    def __init__(self, connections, name):
        self.fs_conn = connections
        self.connections = connections.connections
        self.es = True
        if self.connections['es'] is None:
            self.es = False
        self.name = name
        self.extension = ".json"
        self.kwargs = {}

    @staticmethod
    def dumps_json(d):
        """ Dumps JSON in a central method, so we can clear issues like datetime serialization by passing
            the default parameter globally
        """
        return json.dumps(d, default=str)

    def set_prefix(self, foursight_prefix):
        self.prefix = foursight_prefix

    def get_s3_object(self, key):
        """
        Returns None if not present, otherwise returns a JSON parsed res.
        """
        return self.connections['s3'].get_object(key)

    def get_es_object(self, key):
        """
        Grabs an object from ES. Returns none if not present, json otherwise
        """
        if not self.es:
            return None
        return self.connections['es'].get_object(key)

    def get_object(self, key):
        """
        Gets an object given a key from the data store.
        """
        return self.fs_conn.get_object(key)

    def put_object(self, key, value):
        """
        Puts an object into the data stores
        """
        return self.fs_conn.put_object(key, value)

    def get_latest_result(self):
        """
        Returns the latest result (the last check run)
        """
        latest_key = ''.join([self.name, '/latest', self.extension])
        return self.get_object(latest_key)

    def get_primary_result(self):
        """
        Returns the most recent primary result run (with 'primary'=True in kwargs)
        """
        primary_key = ''.join([self.name, '/primary', self.extension])
        return self.get_object(primary_key)

    def get_result_by_uuid(self, uuid):
        """
        Returns result if it can be found by its uuid, otherwise None.
        """
        result_key = ''.join([self.name, '/', uuid, self.extension])
        return self.get_object(result_key)

    def get_all_results(self):
        """
        Return all results for this check. Should use with care
        """
        all_results = []
        relevant_checks = self.list_keys(records_only=True, prefix=self.name)
        for n in range(len(relevant_checks)):
            check = relevant_checks[n]
            if check.startswith(self.name) and check.endswith(self.extension):
                result = self.get_object(check)
                all_results.append(result)
        return all_results

    def get_closest_result(self, diff_hours=0, diff_mins=0, override_date=None):
        """
        Returns check result that is closest to the current time minus
        diff_hours and diff_mins (both integers).

        If override_date is provided, ignore other arguments and use the given
        date as the metric for finding the check. This MUST be a datetime obj.

        TODO: Add some way to control which results are returned by kwargs?
        For example, you might only want primary results.
        """
        # check_tuples is a list of items of form (s3key, datetime uuid)
        check_tuples = []
        relevant_checks = self.list_keys(records_only=True, prefix=self.name)
        if not relevant_checks:
            raise Exception('Could not find any results for prefix: %s' % self.name)
        # now use only s3 objects with a valid uuid
        for check in relevant_checks:
            if check.startswith(self.name) and check.endswith(self.extension):
                time_str = check[len(self.name) + 1:-len(self.extension)]
                try:
                    check_time = datetime.datetime.strptime(time_str, "%Y-%m-%dT%H:%M:%S.%f")
                except ValueError:
                    continue
                check_time = check_time.replace(tzinfo=tz.tzutc())  # always UTC
                check_tuples.append((check, check_time))
        if override_date:
            desired_time = override_date.replace(tzinfo=tz.tzutc())
        else:
            desired_time = (
                    datetime.datetime.utcnow() - datetime.timedelta(hours=diff_hours, minutes=diff_mins)
            ).replace(tzinfo=tz.tzutc())
        best_match = get_closest(check_tuples, desired_time)
        # ensure that the does not have status 'ERROR'
        match_res = None
        tries = 0  # keep track of number of times we've found an ERROR response
        while not match_res:
            if not best_match or tries > 999:
                raise Exception('Could not find closest non-ERROR result for prefix: %s. Attempted'
                                ' with %s diff hours and %s diff mins.' % (self.name, diff_hours, diff_mins))
            possible_res = self.get_object(best_match[0])
            if possible_res and possible_res.get('status', 'ERROR') != 'ERROR':
                match_res = possible_res
            else:
                check_tuples.remove(best_match)
                if check_tuples:
                    best_match = get_closest(check_tuples, desired_time)
                else:
                    best_match = None
                tries += 1
        return match_res

    def list_keys(self, records_only=True, prefix=None):
        """
        Lists all keys. If given a prefix only keys with that prefix will be
        returned.
        """
        if self.es:
            if prefix:
                return self.connections['es'].list_all_keys_w_prefix(prefix)
            else:
                return self.connections['es'].list_all_keys()
        else:
            if prefix:
                return self.connections['s3'].list_all_keys_w_prefix(prefix, records_only=records_only)
            else:
                return self.connections['s3'].list_all_keys()

    def record_run_info(self):
        """
        Add a record of the completed check to the runs bucket with name
        equal to the dependency id. The object itself is only the status of the run.
        Returns True on success, False otherwise

        NOTE: This functionality is sort of useless I'd say - we don't create a special bucket
        for it anyway, so I am disabling it and replacing the run info with log statement that will
        propagate in the check runner - so you can always search for record_run_info to find these. - Will Nov 15 2022
        """
        run_id = self.kwargs['_run_info']['run_id']
        if not hasattr(self, 'prefix'):
            raise MissingFoursightPrefixException("foursight prefix must be defined using set_prefix")
        # s3_connection = S3Connection(self.prefix + '-runs')
        record_key = '/'.join([run_id, self.name])
        # resp = s3_connection.put_object(record_key, json.dumps(self.status))
        logger.error(f'Skipping record_run_info call to s3: {record_key} -> {run_id}')
        return True  # always returning True now

    def delete_results(self, prior_date=None, primary=True, custom_filter=None, timeout=None, es_only=False):
        """
        Goes through all check files deleting by default all non-primary
        checks. If a prior_date (datetime) is given then all results prior to the
        given time will be delete (including primaries). If primary is False then
        primary results will be cleaned as well.
        If a custom filter is given, that filter will be applied as well, prior
        to the above filters. Note that this argument can be a function or a lambda
        If es_only is specified, only delete the check from ES
        Returns a pair of the number of results deleted from s3 and es respectively
        """

        # since we fall back to s3 get the keys to delete from here
        keys_to_delete = self.connections['s3'].list_all_keys_w_prefix(self.name, records_only=True)

        # if given a custom filter, apply it
        if custom_filter is not None:
            try:
                keys_to_delete = list(filter(custom_filter, keys_to_delete))
            except Exception as e:
                raise Exception('delete_results encountered an error when applying'
                                ' a custom filter. Error message: %s', e)

        # if given a prior date, remove all keys after that date so long as they aren't primary
        if prior_date is not None:
            keys_to_delete = list(filter(lambda k: self.filename_to_datetime(k) <= prior_date, keys_to_delete))

        # if primary is true, filter out primary results (so they arent deleted)
        if primary:
            def is_not_primary(key):
                obj = self.get_s3_object(key)
                return not obj['kwargs'].get('primary')
            keys_to_delete = list(filter(is_not_primary, keys_to_delete))

        # if there is nothing to delete return 0 instead of throwing an Exception
        # in botocore
        if len(keys_to_delete) == 0:
            return 0, 0

        # batch delete calls at aws maximum of 1000 if necessary
        # if timeout is given and reached, return how many we did
        num_deleted_s3, num_deleted_es = 0, 0
        t0 = time.time()
        for i in range(0, len(keys_to_delete), 1000):
            if timeout and round(time.time() - t0, 2) > timeout:
                return num_deleted_s3, num_deleted_es
            if self.es:
                num_deleted_es += self.connections['es'].delete_keys(keys_to_delete[i:i+1000])
            if not es_only:
                s3_resp = self.connections['s3'].delete_keys(keys_to_delete[i:i+1000])
                num_deleted_s3 += len(s3_resp['Deleted'])

        return num_deleted_s3, num_deleted_es

    def get_result_history(self, start, limit, sort='timestamp.desc', after_date=None) -> [list, int]:
        """
        Used to get the uuid, status, and kwargs for a specific check.
        Results are ordered by uuid (timestamped) and sliced from start to limit.
        Probably only called from app_utils.get_foursight_history.
        after_date is an optional datetime object, if provided only the history
        results after that point will be returned.
        Returns a list of lists (inner lists: [status, kwargs])
        """
        if self.es:
            history, total = self.connections['es'].get_result_history(self.name, start, limit, sort)

            def wrapper(obj):
                filename = obj.get('_id')
                if filename is None:
                    filename = obj.get('name') + '/' + obj.get('kwargs').get('uuid') + '.json'
                return self.filename_to_datetime(filename), total

            if after_date is not None:
                history = list(filter(lambda k: wrapper(k)[0] >= after_date, history))
        else:
            all_keys = self.connections['s3'].list_all_keys_w_prefix(self.name, records_only=True)
            total = len(all_keys)
            all_keys.sort(key=self.filename_to_datetime, reverse=True)
            if after_date is not None:
                all_keys = list(filter(lambda k: self.filename_to_datetime(k) >= after_date, all_keys))
            history = all_keys[start:start+limit]
            for idx, key in enumerate(history):
                history[idx] = self.get_s3_object(key)

        results = []
        for n in range(len(history)):
            obj = history[n]
            # order: status <str>, summary <str>, kwargs <dict>, is check? (boolean)
            # handle records that might be malformed
            res_val = [
                obj.get('status', 'Not found'),
                obj.get('summary', None),
                obj.get('kwargs', {}),
                obj.get('type') == 'check' or 'full_output' in obj
            ]
            # kwargs to remove from the history results. these will not be displayed
            for remove_key in ['_run_info']:
                res_val[2].pop(remove_key, None)
            results.append(res_val)
        return results, total

    def store_formatted_result(self, uuid, formatted, primary):
        """
        Store the result in s3/ES. Always makes an entry with key equal to the
        uuid timestamp. Will also store under (i.e. overwrite)the 'latest' key.
        If is_primary, will also overwrite the 'primary' key.

        NOTE: id_alias is an alias of the _id field, which is not searchable in ES >5, which breaks the main page.
        """
        time_key = ''.join([self.name, '/', uuid, self.extension])
        latest_key = ''.join([self.name, '/latest', self.extension])
        primary_key = ''.join([self.name, '/primary', self.extension])

        # store the timestamped result
        formatted['id_alias'] = time_key
        self.put_object(time_key, self.dumps_json(formatted))

        # put result as 'latest' key
        formatted['id_alias'] = latest_key
        self.put_object(latest_key, self.dumps_json(formatted))

        # if primary, store as the primary result
        if primary:
            formatted['id_alias'] = primary_key
            self.put_object(primary_key, self.dumps_json(formatted))
        # return stored data in case we're interested
        return formatted

    def filename_to_datetime(self, key):
        """
        Utility function.
        Key might look like `sync_google_analytics_data/2018-10-15T19:08:32.734656.json`
        We presume that timezone info is not important to allow us to use strptime.
        """
        prefixlen = len(self.name) + 1
        keydatestr = key[prefixlen:-5]  # Remove prefix and .json from key.
        return datetime.datetime.strptime(keydatestr, '%Y-%m-%dT%H:%M:%S.%f')

    @abstractmethod
    def validate(self):
        """
        Validation method that must be implemented by the subclass
        """
        raise NotImplementedError


class CheckResult(RunResult):
    """
    Inherits from RunResult and is meant to be used with checks.

    Usage:
    check = CheckResult(connection, <name>)
    check.status = ...
    check.descritpion = ...
    check.store_result()
    """
    def __init__(self, connections, name, init_uuid=None):
        super().__init__(connections, name)
        # init_uuid arg used if you want to initialize using an existing check
        # init_uuid is in the stringified datetime format
        if init_uuid:
            # maybe make this '.json' dynamic with parent's self.extension?
            ts_key = ''.join([name, '/', init_uuid, '.json'])
            stamp_res = self.get_object(ts_key)
            if stamp_res:
                for key, val in stamp_res.items():
                    if key not in ['uuid', 'kwargs']:  # dont copy these
                        setattr(self, key, val)
                return
        # summary will be displayed next to title when set
        self.summary = ''
        self.description = ''
        # valid values are: 'PASS', 'WARN', 'FAIL', 'ERROR', 'IGNORE'
        # start with IGNORE as the default check status
        self.status = 'IGNORE'
        self.brief_output = None
        self.full_output = None
        # admin output is only seen by admins on the UI
        self.admin_output = None
        self.ff_link = ''
        # self.action_name is the function name of the action to link to check
        self.action = ''
        self.allow_action = False  # by default do not allow the action to be run
        # Added prevent_action to allow the allow_action flag to be True, so action *may* *always*
        # be run, but does not automatically run after check; this came up with the access_key_status
        # check, where we always want to allow the action to be manually run, but want to prevent it from
        # running after the check run, unless it is expiring soon; see: access_key_expiration_detection.
        self.prevent_action = False
        self.action_message = 'Are you sure you want to run this action?'

    def format_result(self, uuid):
        # use summary as description if descrip is missing
        if self.summary and not self.description:
            use_description = self.summary
        else:
            use_description = self.description
        return {
            'name': self.name,
            'summary': self.summary,
            'description': use_description,
            'status': self.status.upper(),
            'uuid': uuid,
            'brief_output': self.brief_output,
            'full_output': self.full_output,
            'admin_output': self.admin_output,
            'ff_link': self.ff_link,
            'action': self.action,
            'allow_action': self.allow_action,
            'prevent_action': self.prevent_action,
            'action_message': self.action_message,
            'kwargs': self.kwargs,
            'type': 'check'
        }

    def validate(self):
        """
        Validates this CheckResult against the elasticsearch mapping
        Multiple errors are possible - the 'latest' wrt when it is checked below
        is reported
        """
        error_msg = []
        store_method = getattr(self, 'store_result', None)
        if not callable(store_method):
            error_msg.append('Do not overwrite the store_result method.')

        def _validate(attr, t):
            if not isinstance(attr, t):
                error_msg.append('%s is not of type %s' % (str(attr), str(t)))

        _validate(self.name, str)
        _validate(self.summary, str)
        _validate(self.description, str)
        _validate(self.status, str)
        _validate(self.ff_link, str)
        _validate(self.action, str)
        _validate(self.allow_action, bool)
        _validate(self.kwargs, dict)
        if len(error_msg) > 0:
            raise BadCheckOrAction('\n'.join(error_msg))

    def store_result(self):
        # normalize status, probably not the optimal place to do this
        if self.status.upper() not in ['PASS', 'WARN', 'FAIL', 'ERROR', 'IGNORE']:
            self.status = 'ERROR'
            self.description = 'Malformed status; look at Foursight check definition.'
        # if there's a set uuid field, use that instead of curr utc time
        # kwargs should ALWAYS have uuid, primary, and queue_action
        if 'uuid' not in self.kwargs:
            self.kwargs['uuid'] = datetime.datetime.utcnow().isoformat()
        if 'primary' not in self.kwargs:
            self.kwargs['primary'] = False
        if 'queue_action' not in self.kwargs:
            self.kwargs['queue_action'] = 'Not queued'
        # if this was triggered from the check_runner, store record of the run
        if '_run_info' in self.kwargs and 'run_id' in self.kwargs['_run_info']:
            self.record_run_info()
        formatted = self.format_result(self.kwargs['uuid'])
        is_primary = self.kwargs.get('primary', False) is True
        # if do_not_store is set, just return result without storing in s3
        if self.kwargs.get('do_not_store', False) is True:
            return formatted
        return self.store_formatted_result(self.kwargs['uuid'], formatted, is_primary)


class ActionResult(RunResult):
    """
    Inherits from RunResult and is meant to be used with actions
    """
    def __init__(self, connections, name):
        self.description = ''
        # valid values are: 'DONE', 'FAIL', 'PEND'
        # start with PEND as the default status
        self.status = 'PEND'
        self.output = None
        super().__init__(connections, name)

    def format_result(self, uuid):
        return {
            'name': self.name,
            'description': self.description,
            'status': self.status.upper(),
            'uuid': uuid,
            'output': self.output,
            'kwargs': self.kwargs,
            'type': 'action'
        }

    def validate(self):
        """
        Validates this action result against the elasticsearch mapping
        Multiple errors are possible - the 'latest' wrt when it is checked below
        is reported
        """
        error_msg = []
        store_method = getattr(self, 'store_result', None)
        if not callable(store_method):
            error_msg.append('Do not overwrite the store_result method.')

        def _validate(attr, t):
            if not isinstance(attr, t):
                error_msg.append('%s is not of type %s' % (str(attr), str(t)))

        _validate(self.status, str)
        _validate(self.description, str)
        _validate(self.name, str)
        _validate(self.kwargs, dict)
        if len(error_msg) > 0:
            raise BadCheckOrAction('\n'.join(error_msg))

    def store_result(self):
        # normalize status, probably not the optimal place to do this
        if self.status.upper() not in ['DONE', 'FAIL', 'PEND']:
            self.status = 'FAIL'
            self.description = 'Malformed status; look at Foursight action definition.'
        # kwargs should **always** have uuid
        if 'uuid' not in self.kwargs:
            self.kwargs['uuid'] = datetime.datetime.utcnow().isoformat()
        # if this was triggered from the check_runner, store record of the run
        if '_run_info' in self.kwargs and 'run_id' in self.kwargs['_run_info']:
            self.record_run_info()
        formatted = self.format_result(self.kwargs['uuid'])
        # if do_not_store is set, just return result without storing in s3
        if self.kwargs.get('do_not_store', False) is True:
            return formatted
        # action results are always stored as 'primary' and 'latest' and can be
        # fetched with the get_latest_result method.
        return self.store_formatted_result(self.kwargs['uuid'], formatted, True)

    def get_associated_check_result(self, kwargs):
        """
        Leverage required 'check_name' and 'called_by' kwargs to return
        the check result from the associted check of this action.
        This will throw a KeyError if the kwargs are missing, but that's okay,
        since we want to enforce the new associated check/action model.
        Must pass in the dict kwargs
        """
        assc_name = kwargs['check_name']
        called_by = kwargs['called_by']
        check = CheckResult(self.fs_conn, assc_name)
        return check.get_result_by_uuid(called_by)


# ===== Utility functions for check_result =====

def get_closest(items, pivot):
    """
    Return the item in the list of items closest to the given pivot.
    Items should be given in tuple form (ID, value (to compare))
    Intended primarily for use with datetime objects.
    See: S.O. 32237862
    """
    return min(items, key=lambda x: abs(x[1] - pivot))



import logging
import os
import requests
from dcicutils.function_cache_decorator import function_cache
from dcicutils.misc_utils import get_error_message
from .misc_utils import get_request_domain, is_running_locally

logging.basicConfig()
logger = logging.getLogger(__name__)


# Class to encapsulate the Auth0 configuration
# parameters from the Portal /auth0_config endpoint,
# e.g. from: https://cgap-mgb.hms.harvard.edu/auth0_config?format=json
#         or: https://data.4dnucleome.org/auth0_config?format=json
#  {
#    "title": "Auth0 Config",
#    "auth0Client": "DPxEwsZRnKDpk0VfVAxrStRKukN14ILB",
#    "auth0Domain": "hms-dbmi.auth0.com",
#    "auth0Options": {
#      "auth": {
#        "sso": false,
#        "redirect": false,
#        "responseType": "token",
#        "params": { "scope": "openid email",
#                    "prompt": "select_account" }
#      },
#      "allowedConnections": [ "github", "google-oauth2", "partners" ]
#    }
#  }
#
# The constructor takes a REQUIRED Portal URL as an argument.
# Use get_config_data() get the relevant Auth0 info in cannocial form;
# it is a dictionary containing these properties:
#
# - domain (e.g. "hms-dbmi.auth0.com")
# - client (e.g. "CQxApsZSnKDpk7VfWMzrTtRKykM14JFC")
# - sso (e.g. False)
# - scope (e.g. "openid email")
# - prompt (e.g. "select_account")
# - connections (e.g. ["github", "google-oauth2"])
#
# Note that these Auth0 configuration parameters are NOT environment specific.
# We ASSUME the same credentials/configuration across all environments for a deployment.
#
class Auth0Config:

    # Fallback values only because at least currently (2022-10-18) this is only returning auth0Client:
    # http://cgap-supertest-1972715139.us-east-1.elb.amazonaws.com/auth0_config?format=json
    # This has been FIXED but leaving this here for now just in case.
    # Note that the secret associated with the Auth0 client is in the GAC (ENCODED_AUTH0_SECRET).
    FALLBACK_VALUES = {
        "domain": "hms-dbmi.auth0.com",
        "client": "DPxEwsZRnKDpk0VfVAxrStRKukN14ILB",
        "sso": False,
        "scope": "openid email",
        "prompt": "select_account",
        "connections": ["github", "google-oauth2"]
    }

    def __init__(self, portal_url: str) -> None:
        if not portal_url:
            raise ValueError("Portal URL required for Auth0Config usage.")
        self._portal_url = portal_url
        self._config_url = f"{portal_url}{'/' if not portal_url.endswith('/') else ''}auth0_config?format=json"

    def get_portal_url(self) -> str:
        return self._portal_url

    def get_config_url(self) -> str:
        return self._config_url

    @function_cache(nocache={})
    def get_config_data(self) -> dict:
        """
        Returns relevant info (dictionary) from the Auth0 config URL in canonical form.
        It contains these properties: domain, client, sso, scope, prompt, connections.
        """
        config_raw_data = self.get_config_raw_data()
        domain = config_raw_data.get("auth0Domain") if config_raw_data else None
        # 2023-04-24: Change to get the Auth0 client ID from the GAC rather
        # that from the Portal /auth0_config endpoint; we still et the other
        # non-credential info (e.g. domain, scope) from that endpoint though.
        # This just makes it more consistent, having both those pieces of info
        # come from the same place. Note FYI that for the non-React code we
        # also get both pieces of the Auth0 credentials from the GAC but
        # the other info is currently hardcoded in the Jinja templates.
        # client = config_raw_data.get("auth0Client") if config_raw_data else None
        client = os.environ.get("CLIENT_ID", os.environ.get("ENCODED_AUTH0_CLIENT"))
        options = config_raw_data.get("auth0Options") if config_raw_data else None
        options_auth = options.get("auth") if options else None
        options_auth_params = options_auth.get("params") if options_auth else None
        sso = options_auth.get("sso") if options_auth else None
        scope = options_auth_params.get("scope") if options_auth_params else None
        prompt = options_auth_params.get("prompt") if options_auth_params else None
        connections = options.get("allowedConnections") if options else None
        # The Auth0 config may contain "partners" in the allowedConnections,
        # which is something we don't want, as it causes a generic diplay
        # of yours@example.com option to login in the Auth0 (lock) box.
        if connections and "partners" in connections:
            connections.remove("partners")
        return {
            "client": client or Auth0Config.FALLBACK_VALUES["client"],
            "domain": domain or Auth0Config.FALLBACK_VALUES["domain"],
            "sso": sso or Auth0Config.FALLBACK_VALUES["sso"],
            "scope": scope or Auth0Config.FALLBACK_VALUES["scope"],
            "prompt": prompt or Auth0Config.FALLBACK_VALUES["prompt"],
            "connections": connections or Auth0Config.FALLBACK_VALUES["connections"]
        }

    @function_cache(nocache={})
    def get_config_raw_data(self) -> dict:
        """
        Returns raw data (dictionary) from the Auth0 config URL.
        """
        try:
            return requests.get(self.get_config_url()).json() or {}
        except Exception as e:
            logger.error(f"Exception fetching Auth0 config ({self.get_config_url()}): {get_error_message(e)}")
            return {}

    def get_client(self) -> str:
        return self.get_config_data()["client"]

    def get_domain(self) -> str:
        return self.get_config_data()["domain"]

    @staticmethod
    def get_secret() -> str:
        """
        Returns the Auth0 secret.
        Currently we get this environment variables setup from the GAC; see identity.py.
        """
        return os.environ.get("CLIENT_SECRET", os.environ.get("ENCODED_AUTH0_SECRET"))

    @staticmethod
    def get_callback_url(request: dict) -> str:
        """
        Returns the URL for our authentication callback endpoint.
        Note this callback endpoint is (still) defined in the legacy Foursight routes.py.
        """
        domain = get_request_domain(request)
        # The context (route prefix) for the Auth0 callback is effectively hardcoded at Auth0.
        # but note different for running locally (localhost) and normal server operation.
        context = "/api/" if not is_running_locally(request) else "/"
        headers = request.get("headers", {})
        scheme = headers.get("x-forwarded-proto", "http")
        return f"{scheme}://{domain}{context}callback/?react"



import logging

from dcicutils.common import EnvName, ChaliceStage
from dcicutils.env_manager import EnvManager
from dcicutils.function_cache_decorator import function_cache
from dcicutils.misc_utils import full_class_name
from dcicutils.s3_utils import s3Utils
from typing import Optional, List
from foursight_core.s3_connection import S3Connection
from dcicutils.env_utils import get_foursight_bucket, get_foursight_bucket_prefix, full_env_name, infer_foursight_from_env


logging.basicConfig()
logger = logging.getLogger(__name__)


class Environment(object):

    def __init__(self, foursight_prefix: Optional[str] = None):  # the foursight_prefix argument can go away.

        prefix = get_foursight_bucket_prefix()

        # This consistency check AND INTERIM HEURISTIC ERROR CORRECTION can go away later,
        # but when it does we can also get
        # rid of the argument that is passed, since it should be possible to look up.
        # For now we have made the argument optional so that it can be gradually phased out,
        # but failing to pass it causes a requirement to declare the "foursight_bucket_prefix"
        # in the ecosystem file of the global env bucket. -kmp 22-Jun-2022
        if not prefix:
            if foursight_prefix is not None:
                prefix = foursight_prefix
                logger.error(f"There is no declared foursight bucket prefix."
                             f" The value from a deprecated data flow path, {foursight_prefix!r}, will be used.")
            else:
                raise RuntimeError(f"There is no declared foursight bucket prefix,"
                                   f" and no suitable substitute can be inferred through legacy means.")
        elif foursight_prefix and prefix != foursight_prefix:
            logger.error(f"A value of deprecated argument 'foursight_prefix', {foursight_prefix},"
                         f" was given when initializing an instance of the class {full_class_name(self)},"
                         f" but the value does not match the declared foursight bucket prefix, {prefix}.")

        self.prefix = prefix
        self.s3_connection = S3Connection(self.get_env_bucket_name())
#       self.cached_list_environment_names = None

    def get_env_bucket_name(self) -> Optional[str]:

        bucket_name = EnvManager.global_env_bucket_name()

        # This consistency check can go away later but has been downgraded to a warning after initial testing
        # confirms this approach can work.
        legacy_bucket_name = self.prefix + '-envs'
        if bucket_name != legacy_bucket_name:
            # Note the inconsistency, but don't flag an error.
            logger.warning(f"{full_class_name(self)}.get_env_bucket_name is returning {bucket_name!r},"
                           f" where it previously would have returned {legacy_bucket_name!r}.")

        return bucket_name

    def list_unique_environment_names(self) -> List[EnvName]:
        result = set()
        for env in self.list_environment_names():
            result.add(infer_foursight_from_env(envname=env))
        return sorted(result)  # a list and sorted

    @function_cache
    def list_environment_names(self) -> List[EnvName]:
        """
        Lists all environments in the foursight-envs s3.

        Returns: a list of names
        """
#       if self.cached_list_environment_names:
#           return self.cached_list_environment_names

        environment_names = [infer_foursight_from_env(envname=env)
                             for env in sorted(EnvManager.get_all_environments(env_bucket=self.get_env_bucket_name()))]

        # This consistency check can go away later..
        legacy_return_value = sorted([key
                                      for key in self.s3_connection.list_all_keys()
                                      if not key.endswith(".ecosystem")])
        modern_full_names = list(map(full_env_name, environment_names))
        legacy_full_names = list(map(full_env_name, legacy_return_value))
        if modern_full_names != legacy_full_names:
            logger.warning(f"{full_class_name(self)}.list_environment_names has consistency problems.")

#       self.cached_list_environment_names = environment_names
        return environment_names

    def list_valid_schedule_environment_names(self) -> List[EnvName]:
        """
        Lists all valid environ names used in schedules including 'all'.

        Returns: A list of names.
        """

        return self.list_unique_environment_names() + ['all']

    def is_valid_environment_name(self, env: Optional[EnvName], or_all: bool = False, strict: bool = False) -> bool:
        """
        Returns True if env is a valid environment name, and False otherwise.

        :param env: The name of an environment.
        :param or_all: if True, allows 'all' as a valid name, otherwise does not.
        :param strict: if True, restricts valid names to exactly the set of foursight names.
                       Otherwise allows all declared names.
        """
        if or_all and env == 'all':
            return True
        valid_envs = self.list_unique_environment_names() if strict else self.list_environment_names()
        return infer_foursight_from_env(envname=env) in valid_envs

    @classmethod
    def get_environment_info_from_s3(cls, env_name: EnvName) -> dict:
        return s3Utils.get_synthetic_env_config(env_name)

    def get_environment_and_bucket_info(self, env_name: EnvName, stage: ChaliceStage) -> dict:

        env_info = self.get_environment_info_from_s3(env_name)

        portal_url = env_info['fourfront']
        es_url = env_info['es']
        ff_env = env_info['ff_env']

        bucket_name = get_foursight_bucket(envname=env_name, stage=stage)

        # This consistency check can go away later. -kmp 22-Jun-2022
        legacy_bucket_name = ''.join([self.prefix + '-', stage, '-', env_name])
        if bucket_name != legacy_bucket_name:
            logger.warning(f"{full_class_name(self)}.get_environment_and_bucket_info({env_name!r}, {stage!r})"
                           f" is returning {bucket_name!r},"
                           f" but it previously would have returned {legacy_bucket_name!r}.")

        env_and_bucket_info = {
            'fourfront': portal_url,
            'es': es_url,
            'ff_env': ff_env,
            'bucket': bucket_name,
        }
        return env_and_bucket_info

    def get_selected_environment_names(self, env_name: EnvName):

        # This is weirdly named. A better name would be expand_environment_names or get_matching_environment_names.
        # But it doesn't need to change. -kmp 24-May-2022
        if env_name == 'all':
            return self.list_unique_environment_names()
        elif self.is_valid_environment_name(env_name):
            return [env_name]
        else:
            raise ValueError(f"Not a valid env name: {env_name}")

    def get_environment_and_bucket_info_in_batch(self, stage: ChaliceStage, env: Optional[EnvName] = None,
                                                 envs: Optional[List[EnvName]] = None):
        """
        Generate environment information from the envs bucket in s3.
        Returns a dictionary keyed by environment name with value of a sub-dict
        with the fields needed to initiate a connection.

        :param stage: a chalice stage (generally one of 'dev' or 'prod')
        :param env: allows you to specify a single env to be initialized, or the token 'all'.
        :param env: allows you to specify a single env to be initialized
        :param envs: allows you to specify multiple envs to be initialized
        """
        if envs is not None:
            if env is not None:
                ValueError(f"{full_class_name(self)}.get_environment_and_bucket_info_in_batch"
                           f" accepts either 'env=' or 'envs=' but not both.")
            env_keys = envs
        else:
            try:
                env_keys = self.get_selected_environment_names(env or 'all')
            except Exception as e:
                logger.warning(f"Returning {{}} from get_environment_and_bucket_info_in_batch due to error: {e}.")
                return {}  # provided env is not in s3

        return {
            env_key: self.get_environment_and_bucket_info(env_key, stage)
            for env_key in env_keys
        }



from dcicutils.ecs_utils import ECSUtils
from dcicutils.env_utils import is_cgap_env
from .helpers.confchecks import *


@check_function()
def ecs_status(connection, **kwargs):
    """ ECS Status check reports metadata on the Clusters/Services running on
        ECS in the account where foursight has been orchestrated.
    """
    check = CheckResult(connection, 'ecs_status')
    full_output = {
        'ECSMeta': {
            'clusters': {}
        }
    }
    client = ECSUtils()
    cluster_arns = client.list_ecs_clusters()
    for cluster_arn in cluster_arns:
        # TODO: When dcicutil 6.7.1 is released replace next line with:
        # if is_c4_arn(cluster_arn):
        # And add at top:
        # from dcicutils.misc_utils import is_c4_arn
        if 'cgap' in cluster_arn or 'fourfront' in cluster_arn:
            cluster_services = client.list_ecs_services(cluster_name=cluster_arn)
            full_output['ECSMeta']['clusters'][cluster_arn] = {
                'services': cluster_services
            }
    if not full_output['ECSMeta']['clusters']:
        check.status = 'WARN'
        check.summary = 'No clusters detected! Has ECS been orchestrated?'
    else:
        check.status = 'PASS'
        check.summary = 'See full output for ECS Metadata'
    check.full_output = full_output
    return check


@check_function(cluster_name=None)
def update_ecs_application_versions(connection, **kwargs):
    """ This check is intended to be run AFTER the user has finished pushing
        the relevant images to ECR. Triggers an update on all services for
        the CGAP cluster. If no cluster_name is passed, Foursight will infer
        one if there is only a single option - otherwise error is raised.

        Note that this check just kicks the process - it does not block until
        the cluster update has finished.
    """
    check = CheckResult(connection, 'update_ecs_application_versions')
    client = ECSUtils()
    cluster_name = kwargs.get('cluster_name')
    cluster_arns = client.list_ecs_clusters()
    env = connection.ff_env
    if is_cgap_env(env):  # just in case we have a fourfront and a cgap
        identifier = 'cgap'
    else:
        identifier = 'fourfront'
    if not cluster_name:
        app_candidate = list(filter(lambda arn: identifier in arn.lower(), cluster_arns))
        if not app_candidate:
            check.status = 'FAIL'
            check.summary = 'No clusters could be resolved from %s' % cluster_arns
        elif len(app_candidate) > 1:
            check.status = 'FAIL'
            check.summary = 'Ambiguous cluster setup (not proceeding): %s' % app_candidate
        else:
            client.update_all_services(cluster_name=app_candidate[0])
            check.status = 'PASS'
            check.summary = 'Triggered cluster update for %s - updating all services.' % app_candidate[0]
    else:
        if cluster_name not in cluster_arns:
            check.status = 'FAIL'
            check.summary = 'Given cluster name does not exist! Gave: %s, Resolved: %s' % (cluster_name, cluster_arns)
        else:
            client.update_all_services(cluster_name=cluster_name)
            check.status = 'PASS'
            check.summary = 'Triggered cluster update for %s - updating all services.' % cluster_name
    return check
